/**************************************************************************
*  This file is part of the TAL project (Tiny Abstraction Layer)
*
*  Copyright (c) 2016 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  27.02.2016  mifi  First Version, tested with an Embedded Artists 
*                    LPC1788 Developer's Kit.
**************************************************************************/
#define __TALCPU_CAN_C__

/*=======================================================================*/
/*  Include                                                              */
/*=======================================================================*/
#include <string.h>
#include "tal.h"

extern void clearAFLUT(LPC_CANAF_T *pCanAF, LPC_CANAF_RAM_T *pCanAFRam);

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

typedef struct          /* Central CAN structure */
{
   __I  uint32_t TXSR;  /* CAN Central Transmit Status Register */
   __I  uint32_t RXSR;  /* CAN Central Receive Status Register */
   __I  uint32_t MSR;   /* CAN Central Miscellaneous Register */
} LPC_CANCR_T;

#define CAN_RXSR_RB1    (1<<8)
#define CAN_RXSR_RB2    (1<<9)
#define CAN_RXSR_DOS1   (1<<16)
#define CAN_RXSR_DOS2   (1<<17)

#define CAN_MSR_E1      (1<<0)
#define CAN_MSR_E2      (1<<1)

#define CAN_SJW_1tq     0x00
#define CAN_SJW_2tq     0x01
#define CAN_SJW_3tq     0x02
#define CAN_SJW_4tq     0x03

#define CAN_TSEG1_1tq   0x00
#define CAN_TSEG1_2tq   0x01
#define CAN_TSEG1_3tq   0x02
#define CAN_TSEG1_4tq   0x03
#define CAN_TSEG1_5tq   0x04
#define CAN_TSEG1_6tq   0x05
#define CAN_TSEG1_7tq   0x06
#define CAN_TSEG1_8tq   0x07
#define CAN_TSEG1_9tq   0x08
#define CAN_TSEG1_10tq  0x09
#define CAN_TSEG1_11tq  0x0A
#define CAN_TSEG1_12tq  0x0B
#define CAN_TSEG1_13tq  0x0C
#define CAN_TSEG1_14tq  0x0D
#define CAN_TSEG1_15tq  0x0E
#define CAN_TSEG1_16tq  0x0F

#define CAN_TSEG2_1tq   0x00
#define CAN_TSEG2_2tq   0x01
#define CAN_TSEG2_3tq   0x02
#define CAN_TSEG2_4tq   0x03
#define CAN_TSEG2_5tq   0x04
#define CAN_TSEG2_6tq   0x05
#define CAN_TSEG2_7tq   0x06
#define CAN_TSEG2_8tq   0x07

   
/*
 * Some CAN timing calculators:
 * http://www.bittiming.can-wiki.info
 * http://www.port.de/de/bit-timing.html
 */ 

/*
 * BTR defines for 60MHz clock, Sample Point 86.7% (15tq)
 */
#define BTR_1000K_60MHZ ((CAN_SJW_1tq<<14) | (CAN_TSEG2_2tq<<20) | (CAN_TSEG1_12tq<<16) | (  4-1))
#define BTR_500K_60MHZ  ((CAN_SJW_1tq<<14) | (CAN_TSEG2_2tq<<20) | (CAN_TSEG1_12tq<<16) | (  8-1))
#define BTR_250K_60MHZ  ((CAN_SJW_1tq<<14) | (CAN_TSEG2_2tq<<20) | (CAN_TSEG1_12tq<<16) | ( 16-1))
#define BTR_125K_60MHZ  ((CAN_SJW_1tq<<14) | (CAN_TSEG2_2tq<<20) | (CAN_TSEG1_12tq<<16) | ( 32-1))
#define BTR_100K_60MHZ  ((CAN_SJW_1tq<<14) | (CAN_TSEG2_2tq<<20) | (CAN_TSEG1_12tq<<16) | ( 40-1))
#define BTR_50K_60MHZ   ((CAN_SJW_1tq<<14) | (CAN_TSEG2_2tq<<20) | (CAN_TSEG1_12tq<<16) | ( 80-1))
#define BTR_20K_60MHZ   ((CAN_SJW_1tq<<14) | (CAN_TSEG2_2tq<<20) | (CAN_TSEG1_12tq<<16) | (200-1))
#define BTR_10K_60MHZ   ((CAN_SJW_1tq<<14) | (CAN_TSEG2_2tq<<20) | (CAN_TSEG1_12tq<<16) | (400-1))
#define BTR_5K_60MHZ    ((CAN_SJW_1tq<<14) | (CAN_TSEG2_2tq<<20) | (CAN_TSEG1_12tq<<16) | (800-1))


/*
 * Filter defines
 */
#define FILTER_ID_MASK        0x1FFFFFFF
#define FILTER_IN_USE_FLAG    0x80000000
#define FILTER_IS_EXT_FLAG    0x40000000

#define FILTER_ID(_x)         (_x & FILTER_ID_MASK)
#define IS_FILTER_IN_USE(_x)  (_x & FILTER_IN_USE_FLAG)
#define IS_FILTER_EXT(_x)     (_x & FILTER_IS_EXT_FLAG)

#define STAT_INC(_x)          if (_x < UINT32_MAX) _x++


/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

/*
 * Device Control Block 
 */
static TAL_CAN_DCB   *DCBArray[TAL_CAN_PORT_MAX];

/*
 * Acceptance Filter
 */
static CAN_STD_ID_ENTRY_T       AFStdSection[TAL_CAN_PORT_MAX * CAN_MAX_FILTER_COUNT];
static CAN_EXT_ID_ENTRY_T       AFExtSection[TAL_CAN_PORT_MAX * CAN_MAX_FILTER_COUNT];
static CAN_STD_ID_RANGE_ENTRY_T AFStdGroupSection[TAL_CAN_PORT_MAX];
static CAN_EXT_ID_RANGE_ENTRY_T AFExtGroupSection[TAL_CAN_PORT_MAX];
static CANAF_LUT_T              AFSections;

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*************************************************************************/
/*  SortFilterList                                                       */
/*                                                                       */
/*  Sort the filter list in ascending numerical order.                   */
/*                                                                       */
/*  In    : Filter, FilterSize                                           */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void SortFilterList (uint32_t *Filter, uint16_t FilterSize)
{
   uint32_t TmpFilter;

   uint16_t i;
   uint16_t j;
   
   for (i = 0; i < FilterSize - 1; i++)
   {
      for (j = 0; j < (FilterSize - i) - 1; j++)
      {
         if (FILTER_ID(Filter[j]) > FILTER_ID(Filter[j + 1]))
         {
            TmpFilter     = Filter[j];
            Filter[j]     = Filter[j + 1];
            Filter[j + 1] = TmpFilter;
         }
      }
   }
} /* SortFilterList */

/*************************************************************************/
/*  UpdateAcceptanceFilter                                               */
/*                                                                       */
/*  Update the acceptance filter.                                        */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
static TAL_RESULT UpdateAcceptanceFilter (void)
{
   TAL_RESULT    Error = TAL_ERROR;
   TAL_CAN_DCB *pDCB;
   uint8_t      bPort;
   uint16_t     wIndex;
   uint16_t     wStdCount      = 0;
   uint16_t     wExtCount      = 0;
   uint16_t     wStdGroupCount = 0;
   uint16_t     wExtGroupCount = 0;
   Status       CANStatus;

   /* STD section */    
   for (bPort = 0; bPort < TAL_CAN_PORT_MAX; bPort++)
   {
      pDCB = DCBArray[bPort];
      if (pDCB != NULL)
      {
         SortFilterList(pDCB->HW.FilterArray, CAN_MAX_FILTER_COUNT); 
   
         for (wIndex = 0; wIndex < CAN_MAX_FILTER_COUNT; wIndex++)
         {
            if (IS_FILTER_IN_USE(pDCB->HW.FilterArray[wIndex]))
            {
               if (IS_FILTER_EXT(pDCB->HW.FilterArray[wIndex]))
               {
                  AFExtSection[wExtCount].CtrlNo  = bPort;
                  AFExtSection[wExtCount].ID_29   = FILTER_ID(pDCB->HW.FilterArray[wIndex]);
                  
                  wExtCount++;
               }
               else
               {
                  AFStdSection[wStdCount].CtrlNo  = bPort;
                  AFStdSection[wStdCount].Disable = 0;
                  AFStdSection[wStdCount].ID_11   = (uint16_t)FILTER_ID(pDCB->HW.FilterArray[wIndex]);
                  
                  wStdCount++;
               }   
            }
         }
         
         if (TAL_TRUE == pDCB->HW.bFilterAllStd)
         {
            AFStdGroupSection[wStdGroupCount].LowerID.CtrlNo  = bPort;
            AFStdGroupSection[wStdGroupCount].LowerID.Disable = 0;
            AFStdGroupSection[wStdGroupCount].LowerID.ID_11   = 0;
            
            AFStdGroupSection[wStdGroupCount].UpperID.CtrlNo  = bPort;
            AFStdGroupSection[wStdGroupCount].UpperID.Disable = 0;
            AFStdGroupSection[wStdGroupCount].UpperID.ID_11   = CAN_STD_ENTRY_ID_MASK;
            
            wStdGroupCount++;
         } 

         if (TAL_TRUE == pDCB->HW.bFilterAllExt)
         {
            AFExtGroupSection[wExtGroupCount].LowerID.CtrlNo  = bPort;
            AFExtGroupSection[wExtGroupCount].LowerID.ID_29   = 0;
            
            AFExtGroupSection[wExtGroupCount].UpperID.CtrlNo  = bPort;
            AFExtGroupSection[wExtGroupCount].UpperID.ID_29   = CAN_EXT_ENTRY_ID_MASK;
            
            wExtGroupCount++;            
         } 
         
      } /* end if (pDCB != NULL) */ 
   } /* end for (bPort = 0; bPort < TAL_CAN_PORT_MAX; bPort++) */
   

   /*
    * Set Acceptance Filter sections
    */
   AFSections.FullCANSec      = NULL;
   AFSections.FullCANEntryNum = 0;
   
   AFSections.SffSec          = AFStdSection;
   AFSections.SffEntryNum     = wStdCount;
   
   AFSections.SffGrpSec       = AFStdGroupSection;
   AFSections.SffGrpEntryNum  = wStdGroupCount;
   
   AFSections.EffSec          = AFExtSection;
   AFSections.EffEntryNum     = wExtCount;
   
   AFSections.EffGrpSec       = AFExtGroupSection;
   AFSections.EffGrpEntryNum  = wExtGroupCount;
   
   CANStatus = Chip_CAN_SetAFLUT(LPC_CANAF, LPC_CANAF_RAM, &AFSections);   
   if (SUCCESS == CANStatus)
   {
      Error = TAL_OK;
   }

   return(Error);
} /* UpdateAcceptanceFilter */

/*************************************************************************/
/*  SendObject                                                           */
/*                                                                       */
/*  Try to send the given Object.                                        */
/*                                                                       */
/*  In    : pCANx, pCANObject                                            */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
static TAL_RESULT SendObject (LPC_CAN_T *pCANx, TAL_CAN_OBJECT *pCANObject)
{
   TAL_RESULT  Error;
   CAN_MSG_T   TxMsgBuf;
   
   
   TxMsgBuf.ID   = pCANObject->dIdentifier;
   TxMsgBuf.DLC  = pCANObject->bDLC;
   TxMsgBuf.Type = 0;
   
   if (pCANObject->bFlags & TAL_CAN_OBJECT_FLAGS_EXT_ID)
   {
      TxMsgBuf.ID |= CAN_EXTEND_ID_USAGE;
   }   
   
   TxMsgBuf.Data[0] = pCANObject->Data[0];
   TxMsgBuf.Data[1] = pCANObject->Data[1];
   TxMsgBuf.Data[2] = pCANObject->Data[2];
   TxMsgBuf.Data[3] = pCANObject->Data[3];
   TxMsgBuf.Data[4] = pCANObject->Data[4];
   TxMsgBuf.Data[5] = pCANObject->Data[5];
   TxMsgBuf.Data[6] = pCANObject->Data[6];
   TxMsgBuf.Data[7] = pCANObject->Data[7];
   
   Chip_CAN_Send(pCANx, CAN_BUFFER_1, &TxMsgBuf);

   Error = TAL_OK;

//SendObjectEnd:
   return(Error);   
} /* SendObject */

/*************************************************************************/
/*  TX_IRQHandler                                                        */
/*                                                                       */
/*  This is the generic TX IRQhandler.                                   */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void TX_IRQHandler (TAL_CAN_DCB *pDCB)
{
   TAL_RESULT     Error;
   TAL_CAN_OBJECT CANObject;
   LPC_CAN_T    *pCANx;

   /* Get CAN pointer */
   pCANx = (LPC_CAN_T *)pDCB->HW.dBaseAddress;

   /* 
    * Check for TX interrupt, but only if it is anabled 
    */
   if (pCANx->IER & CAN_IER_TIE1)
   {
      /* Read Object from the ring buffer */
      Error = tal_MISCRingGet(&pDCB->TxRing, (uint8_t*)&CANObject);
      if (Error != TAL_OK)
      {
         /* Ups, no data available, disable interrupts */
         pCANx->IER &= ~(CAN_IER_TIE1 | CAN_IER_TIE2 | CAN_IER_TIE3);
      }
      else
      {
         /* Send Object */
         SendObject(pCANx, &CANObject); 
      }   
   } /* end "TX interrupt */
   
   STAT_INC(pDCB->HW.Stat.dTXCount);
   
} /* TX_IRQHandler */

/*************************************************************************/
/*  RX_IRQHandler                                                        */
/*                                                                       */
/*  This is the generic RX IRQhandler.                                   */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void RX_IRQHandler (TAL_CAN_DCB *pDCB)
{
   TAL_RESULT     Error;
   TAL_CAN_OBJECT CANObject;
   LPC_CAN_T    *pCANx;
   CAN_MSG_T      RcvMsgBuf;
   
   /* Set timestamp, if supported */
   if (NULL == pDCB->GetRxTimestamp)
   { 
      CANObject.dTimestamp = 0;
   }
   else
   {
      CANObject.dTimestamp = pDCB->GetRxTimestamp();
   }   

   /* Get CAN pointer */
   pCANx = (LPC_CAN_T *)pDCB->HW.dBaseAddress;

   Chip_CAN_Receive(pCANx, &RcvMsgBuf);

   /* Get Identifier and EXT/STD info */
   if (RcvMsgBuf.ID & CAN_EXTEND_ID_USAGE)
   {
      CANObject.dIdentifier = CAN_RID_ID_29(RcvMsgBuf.ID);
      CANObject.bFlags      = TAL_CAN_OBJECT_FLAGS_EXT_ID;
   }
   else
   {
      CANObject.dIdentifier = CAN_RID_ID_11(RcvMsgBuf.ID);
      CANObject.bFlags      = TAL_CAN_OBJECT_FLAGS_STD_ID;
   }   
   
   /* Get DLC */
   CANObject.bDLC = (uint8_t)RcvMsgBuf.DLC;
   
   /* Get data */
   CANObject.Data[0] = RcvMsgBuf.Data[0];
   CANObject.Data[1] = RcvMsgBuf.Data[1];
   CANObject.Data[2] = RcvMsgBuf.Data[2];
   CANObject.Data[3] = RcvMsgBuf.Data[3];
   CANObject.Data[4] = RcvMsgBuf.Data[4];
   CANObject.Data[5] = RcvMsgBuf.Data[5];
   CANObject.Data[6] = RcvMsgBuf.Data[6];
   CANObject.Data[7] = RcvMsgBuf.Data[7];
   
   /* If we have no overflow... */
   if (TAL_FALSE == pDCB->bRxOverflow)
   {
      /* ... put it into the ring buffer */
      Error = tal_MISCRingAdd(&pDCB->RxRing, (uint8_t*)&CANObject);
      if (TAL_OK == Error)
      {
         /* Signal counting semaphore */
         OSSemaSignalFromInt(&pDCB->RxRdySema);
      }
      else
      {
         /* Ups, overflow */
         pDCB->bRxOverflow = TAL_OK;
      }
   } 
   
   STAT_INC(pDCB->HW.Stat.dRXCount); 
   
} /* RX_IRQHandler */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  CAN_IRQHandler                                                       */
/*                                                                       */
/*  This is the Cortex CAN IRQ handler.                                  */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void CAN_IRQHandler (void)
{
   uint32_t dIntStatus;
   uint32_t dRxStatus;


   TAL_CPU_IRQ_ENTER();

   /*
    * Check RX status for CAN1 and CAN2
    */   
   dRxStatus = LPC_CANCR->RXSR;
   if (dRxStatus & CAN_RXSR_RB1)
   {
      RX_IRQHandler(DCBArray[TAL_CAN_PORT_1]);
   }
   if (dRxStatus & CAN_RXSR_RB2)
   {
      RX_IRQHandler(DCBArray[TAL_CAN_PORT_2]);
   }
   
      
   /*
    * Check DOS status for CAN1 and CAN2
    */
   if (dRxStatus & CAN_RXSR_DOS1)
   {
      STAT_INC(DCBArray[TAL_CAN_PORT_1]->HW.Stat.dOverunCount);
   }
   if (dRxStatus & CAN_RXSR_DOS2)
   {
      STAT_INC(DCBArray[TAL_CAN_PORT_2]->HW.Stat.dOverunCount);
   }

   
   /*
    * Check TX status for CAN1
    */   
   if (TAL_TRUE == DCBArray[TAL_CAN_PORT_1]->bIsOpen) 
   {    
      dIntStatus = Chip_CAN_GetIntStatus(LPC_CAN1);
      if (dIntStatus & CAN_ICR_TI1)
      {
         TX_IRQHandler(DCBArray[TAL_CAN_PORT_1]);
      }
   }      
   
   /*
    * Check TX status for CAN2
    */   
   if (TAL_TRUE == DCBArray[TAL_CAN_PORT_2]->bIsOpen) 
   {    
      dIntStatus = Chip_CAN_GetIntStatus(LPC_CAN2);
      if (dIntStatus & CAN_ICR_TI1)
      {
         TX_IRQHandler(DCBArray[TAL_CAN_PORT_2]);
      }
   }      
   
   TAL_CPU_IRQ_EXIT();
} /* CAN_IRQHandler */

/*************************************************************************/
/*  cpu_CANInit                                                          */
/*                                                                       */
/*  Prepare the hardware for use by the Open function later. Set the HW  */
/*  information depending of ePort and "enable" the CAN port.            */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_CANInit (TAL_CAN_DCB *pDCB)
{
   TAL_RESULT      Error     = TAL_ERR_CAN_PORT_RANGE;
   TAL_CAN_HW    *pHW        = &pDCB->HW;
   static uint8_t bFirstInit = TRUE;
      
   switch (pDCB->ePort)
   {
      case TAL_CAN_PORT_1:
      {
         Error = tal_BoardEnableCAN1();
         if (TAL_OK == Error)
         {
            DCBArray[TAL_CAN_PORT_1] = pDCB;

            /* Clear HW info first */
            memset(pHW, 0x00, sizeof(TAL_CAN_HW));

            pHW->dBaseAddress = LPC_CAN1_BASE;
            
            /* Reset CAN1 */
            Chip_SYSCTL_PeriphReset(SYSCTL_RESET_CAN1);
         }
         break;
      } /* TAL_CAN_PORT_1 */
   
      case TAL_CAN_PORT_2:
      {
         Error = tal_BoardEnableCAN2();
         if (TAL_OK == Error)
         {
            DCBArray[TAL_CAN_PORT_2] = pDCB;

            /* Clear HW info first */
            memset(pHW, 0x00, sizeof(TAL_CAN_HW));

            pHW->dBaseAddress = LPC_CAN2_BASE;
            
            /* Reset CAN2 */
            Chip_SYSCTL_PeriphReset(SYSCTL_RESET_CAN2);
         }
         break;
      } /* TAL_CAN_PORT_2 */
      
      default:
      {
         (void)LPC_CANCR->TXSR;  /* Prevent lint warning */ 
      
         /* Do nothing */
         Error = TAL_ERR_CAN_PORT_RANGE;
         break;
      }
   } /* end switch (pDCB->ePort) */
   
      
   /*
    * Check if the acceptance filter must be initialized.
    * We have only one big acceptance filter for all the
    * controllers. Therefore must be initialized only one time.
    */
   if ((TAL_TRUE == bFirstInit) && (TAL_OK == Error))
   {
      bFirstInit = TAL_FALSE;

	   /* Initialize Acceptance filter one time */
	   clearAFLUT(LPC_CANAF, LPC_CANAF_RAM);
   }
   
   return(Error);
} /* cpu_CANInit */

/*************************************************************************/
/*  cpu_CANOpen                                                          */
/*                                                                       */
/*  Open the CAN port.                                                   */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_CANOpen (TAL_CAN_DCB *pDCB)
{
   TAL_RESULT  Error = TAL_ERROR;
   LPC_CAN_T *pCANx;
   uint32_t   dBTR;
   
   /* Get CAN pointer */
   pCANx = (LPC_CAN_T *)pDCB->HW.dBaseAddress;

   /* 
    * Check baud rate 
    */
   if (60000000L == tal_CPUGetFrequencyAPB())
   {
      switch (pDCB->Settings.dBitRate)
      {
         case TAL_CAN_BR_1000K: dBTR = BTR_1000K_60MHZ; break;
         case TAL_CAN_BR_500K:  dBTR = BTR_500K_60MHZ;  break;
         case TAL_CAN_BR_250K:  dBTR = BTR_250K_60MHZ;  break;
         case TAL_CAN_BR_125K:  dBTR = BTR_125K_60MHZ;  break;
         case TAL_CAN_BR_100K:  dBTR = BTR_100K_60MHZ;  break;
         case TAL_CAN_BR_50K:   dBTR = BTR_50K_60MHZ;   break;
         case TAL_CAN_BR_20K:   dBTR = BTR_20K_60MHZ;   break;
         case TAL_CAN_BR_10K:   dBTR = BTR_10K_60MHZ;   break;
         case TAL_CAN_BR_5K:    dBTR = BTR_5K_60MHZ;    break;
      
         default:
         {
            Error = TAL_ERR_CAN_BAUDRATE;
            goto CANOpenEnd;  /*lint !e801*/
            break;   /*lint !e527*/
         }
      } /* end switch (pDCB->Settings.dBaudrate) */
   }
   else
   {
      Error = TAL_ERR_CAN_BAUDRATE;
      goto CANOpenEnd;  /*lint !e801*/
   }

   /* Clear filter */   
   memset(pDCB->HW.FilterArray, 0x00, sizeof(pDCB->HW.FilterArray));
   pDCB->HW.bFilterAllStd = TAL_FALSE;
   pDCB->HW.bFilterAllExt = TAL_FALSE;

   /* General initialization */
   Chip_CAN_Init(pCANx, LPC_CANAF, LPC_CANAF_RAM);
   
   /* Set  BitRate */
   Chip_CAN_SetMode(pCANx, CAN_RESET_MODE, ENABLE);
   pCANx->BTR = dBTR;
   Chip_CAN_SetMode(pCANx, CAN_RESET_MODE, DISABLE);

   /* Enable receive interrupt */   
   Chip_CAN_EnableInt(pCANx, CAN_IER_RIE);

   /* Enable some error interrupts */   
   Chip_CAN_EnableInt(pCANx, CAN_IER_DOIE);
   
   /* Set Acceptance Filter Mode */
   Chip_CAN_SetAFMode(LPC_CANAF, CAN_AF_NORMAL_MODE);
   
   /* Set priority and enable the IRQ */
   tal_CPUIrqSetPriority(CAN_IRQn, CAN_PRIO);
   tal_CPUIrqEnable(CAN_IRQn);
   
   Error = TAL_OK;
   
CANOpenEnd:   
   return(Error);
} /* cpu_CANOpen */

/*************************************************************************/
/*  cpu_CANClose                                                         */
/*                                                                       */
/*  Close the CAN port.                                                  */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_CANClose (TAL_CAN_DCB *pDCB)
{
   TAL_RESULT    Error = TAL_OK;
   LPC_CAN_T   *pCANx;
   
   /* Get CAN pointer */
   pCANx = (LPC_CAN_T *)pDCB->HW.dBaseAddress;

   Chip_CAN_DeInit(pCANx);
   
   /* Clear filter and update acceptance filter */   
   memset(pDCB->HW.FilterArray, 0x00, sizeof(pDCB->HW.FilterArray));
   pDCB->HW.bFilterAllStd = TAL_FALSE;
   pDCB->HW.bFilterAllExt = TAL_FALSE;
   
   UpdateAcceptanceFilter();
      
   return(Error);
} /* cpu_CANClose */

/*************************************************************************/
/*  cpu_CANStartTx                                                       */
/*                                                                       */
/*  Send the data from the ring buffer if the TX interrupt is disabled.  */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_CANStartTx (TAL_CAN_DCB *pDCB)
{
   TAL_RESULT       Error = TAL_ERROR;
   LPC_CAN_T      *pCANx;
   TAL_CAN_OBJECT   CANObject;
   
   /* Get CAN pointer */
   pCANx = (LPC_CAN_T *)pDCB->HW.dBaseAddress;

   TAL_CPU_DISABLE_ALL_INTS();
   if (pCANx->IER & CAN_IER_TIE1)
   {
      /* TX interrupt is enabled, do nothing */
   }
   else
   {
      Error = tal_MISCRingGet(&pDCB->TxRing, (uint8_t*)&CANObject);
      if (TAL_OK == Error)
      {
         Error = SendObject(pCANx, &CANObject);
         if (TAL_OK == Error)
         {
            /* Enable TX interrupt */
            pCANx->IER |= CAN_IER_TIE1;
//            pCANx->IER |= (CAN_IER_TIE1 | CAN_IER_TIE2);
//            pCANx->IER |= (CAN_IER_TIE1 | CAN_IER_TIE2 | CAN_IER_TIE3);
//            Chip_CAN_EnableInt(pCANx, CAN_IER_TIE1);
         }
      }
   }
   TAL_CPU_ENABLE_ALL_INTS();
   
   return(Error);
} /* cpu_CANStartTx */

/*************************************************************************/
/*  cpu_CANTxIsRunning                                                   */
/*                                                                       */
/*  Check if TX is still running.                                        */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / TAL_ERROR                                           */
/*************************************************************************/
TAL_RESULT cpu_CANTxIsRunning (TAL_CAN_DCB *pDCB)
{
   TAL_RESULT  Error = TAL_OK;
   LPC_CAN_T *pCANx;
   
   /* Get CAN pointer */
   pCANx = (LPC_CAN_T *)pDCB->HW.dBaseAddress;

   TAL_CPU_DISABLE_ALL_INTS();
   if (pCANx->IER & CAN_IER_TIE1)
   {
      /* TX is still running */
      Error = TAL_OK;
   }
   else
   {
      /* TX is not running */
      Error = TAL_OK;
   }
   TAL_CPU_ENABLE_ALL_INTS();
   
   return(Error);
} /* cpu_CANTxIsRunning */

/*************************************************************************/
/*  cpu_CANEnableSilentMode                                              */
/*                                                                       */
/*  Enable "Silent" mode.                                                */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_CANEnableSilentMode (TAL_CAN_DCB *pDCB)
{
   (void)pDCB;   
   
   /* Not supported */
   
   return(TAL_ERROR);
} /* cpu_CANEnableSilentMode */

/*************************************************************************/
/*  cpu_CANDisableSilentMode                                             */
/*                                                                       */
/*  Disable "Silent" mode.                                               */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_CANDisableSilentMode (TAL_CAN_DCB *pDCB)
{
   (void)pDCB;   
   
   /* Not supported */
   
   return(TAL_ERROR);
} /* cpu_CANDisableSilentMode */

/*************************************************************************/
/*  cpu_CANIdentRegister                                                 */
/*                                                                       */
/*  Register a CAN identifier.                                           */
/*                                                                       */
/*  In    : pDCB, dIdentifier, Type                                      */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_CANIdentRegister (TAL_CAN_DCB  *pDCB,
                                uint32_t       dIdentifier,
                                TAL_CAN_TYPE_ID Type)
{
   TAL_RESULT  Error = TAL_ERROR;
   uint16_t   wIndex;
   uint32_t  *FilterArray = pDCB->HW.FilterArray;
   uint8_t    bRegsiterOK = TAL_FALSE;
   
   
   /* 
    * Handle special case TAL_CAN_ID_RX_ALL, all identifier
    * from type STD or EXT should be received.
    */
   if (TAL_CAN_ID_RX_ALL == dIdentifier)
   {
      /* Test if STD or EXT ID */
      if (TAL_CAN_TYPE_STD_ID == Type)
      {
         pDCB->HW.bFilterAllStd = TAL_TRUE;
      }
      else
      {
         pDCB->HW.bFilterAllExt = TAL_TRUE;
      }   

      Error = UpdateAcceptanceFilter();    
      goto IdentRegisterEnd;     /*lint !e801*/
   } /* end if (TAL_CAN_ID_RX_ALL == dIdentifier) */


   /*
    * Check for STD or EXT identifier
    */   
   if (TAL_CAN_TYPE_STD_ID == Type)
   {
      /* Check valid range */
      if (dIdentifier > CAN_STD_ENTRY_ID_MASK)
      {
         Error = TAL_ERR_CAN_ID_RANGE;
         goto IdentRegisterEnd;  /*lint !e801*/
      }
   }
   else
   {
      /* Check valid range */
      if (dIdentifier > CAN_EXT_ENTRY_ID_MASK)
      {
         Error = TAL_ERR_CAN_ID_RANGE;
         goto IdentRegisterEnd;  /*lint !e801*/
      }

      dIdentifier |= FILTER_IS_EXT_FLAG; 
   }
   
   dIdentifier |= FILTER_IN_USE_FLAG;
   
   /*
    * Check if this identifier is registered
    */
   for (wIndex = 0; wIndex < CAN_MAX_FILTER_COUNT; wIndex++)
   {
      if (FilterArray[wIndex] == dIdentifier)
      {
         Error = TAL_ERR_CAN_ID_USED;
         goto IdentRegisterEnd;  /*lint !e801*/
      }
   }
   
   
   /*
    * At this point the identifier is not registered.
    * Find a free entry which can be used.
    */
   for (wIndex = 0; wIndex < CAN_MAX_FILTER_COUNT; wIndex++)
   {
      if (0 == IS_FILTER_IN_USE(FilterArray[wIndex]))
      {
         FilterArray[wIndex] = dIdentifier;
         bRegsiterOK         = TAL_TRUE;
         break;
      }
   }      
   if (TAL_FALSE == bRegsiterOK)
   {
      Error = TAL_ERR_CAN_ID_NO_ENTRY;
      goto IdentRegisterEnd;  /*lint !e801*/
   }


   /*
    * At this point an update of the Acceptance Filter is needed
    */
   Error = UpdateAcceptanceFilter();    
   
IdentRegisterEnd:   
   return(Error);
} /* cpu_CANIdentRegister */   

/*************************************************************************/
/*  cpu_CANIdentDeRegister                                               */
/*                                                                       */
/*  DeRegister a CAN identifier.                                         */
/*                                                                       */
/*  In    : pDCB, dIdentifier, Type                                      */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_CANIdentDeRegister (TAL_CAN_DCB  *pDCB,
                                  uint32_t       dIdentifier,
                                  TAL_CAN_TYPE_ID Type)
{
   TAL_RESULT  Error = TAL_ERROR;     
   uint16_t   wIndex;
   uint32_t  *FilterArray   = pDCB->HW.FilterArray;
   uint8_t    bDeRegsiterOK = TAL_FALSE;
  

   /* 
    * Handle special case TAL_CAN_ID_RX_ALL, all identifier
    * from type STD or EXT should be received.
    */
   if (TAL_CAN_ID_RX_ALL == dIdentifier)
   {
      /* Test if STD or EXT ID */
      if (TAL_CAN_TYPE_STD_ID == Type)
      {
         pDCB->HW.bFilterAllStd = TAL_FALSE;
      }
      else
      {
         pDCB->HW.bFilterAllExt = TAL_FALSE;
      }   

      Error = UpdateAcceptanceFilter();    
      goto IdentDeRegisterEnd;      /*lint !e801*/
   } /* end if (TAL_CAN_ID_RX_ALL == dIdentifier) */

  
   /*
    * Check for STD or EXT identifier
    */   
   if (TAL_CAN_TYPE_STD_ID == Type)
   {
      /* Check valid range */
      if (dIdentifier > CAN_STD_ENTRY_ID_MASK)
      {
         Error = TAL_ERR_CAN_ID_RANGE;
         goto IdentDeRegisterEnd;   /*lint !e801*/
      }
   }
   else
   {
      /* Check valid range */
      if (dIdentifier > CAN_EXT_ENTRY_ID_MASK)
      {
         Error = TAL_ERR_CAN_ID_RANGE;
         goto IdentDeRegisterEnd;   /*lint !e801*/
      }
      
      dIdentifier |= FILTER_IS_EXT_FLAG; 
   }
   
   dIdentifier |= FILTER_IN_USE_FLAG;
   
   /*
    * Check if this identifier is registered,
    * and remove it from the list if available.
    */
   for (wIndex = 0; wIndex < CAN_MAX_FILTER_COUNT; wIndex++)
   {
      if (FilterArray[wIndex] == dIdentifier)
      {
         FilterArray[wIndex] = 0;
         bDeRegsiterOK        = TAL_TRUE;
         break;
      }
   }
   if (TAL_FALSE == bDeRegsiterOK)
   {
      Error = TAL_ERR_CAN_ID_NOT_USED;
      goto IdentDeRegisterEnd;   /*lint !e801*/
   }

   /*
    * At this point an update of the Acceptance Filter is needed
    */
   Error = UpdateAcceptanceFilter();    

IdentDeRegisterEnd:   
   return(Error);
} /* cpu_CANIdentDeRegister */   

/*** EOF ***/
